"""A HTTP API transport that offers optional local caching of the results."""

import datetime
import enum
import hashlib
import json
import os
import pathlib
import platform
import re
from contextlib import contextmanager
from importlib.metadata import version
from json import JSONDecodeError
from typing import Optional, Callable, Union, Collection, Dict, Tuple
import shutil
import logging
from pathlib import Path
from urllib3 import Retry

import pandas
import pandas as pd
import requests
from filelock import FileLock
from requests import Response
from requests.adapters import HTTPAdapter

from tradingstrategy.candle import TradingPairDataAvailability
from tradingstrategy.chain import ChainId
from tradingstrategy.liquidity import XYLiquidity
from tradingstrategy.timebucket import TimeBucket
from tradingstrategy.transport.jsonl import load_candles_jsonl
from tradingstrategy.types import PrimaryKey
from tradingstrategy.lending import LendingCandle, LendingCandleType
from tradingstrategy.transport.progress_enabled_download import download_with_tqdm_progress_bar


from tqdm_loggable.auto import tqdm


logger = logging.getLogger(__name__)


class OHLCVCandleType(enum.Enum):
    """Candle types for /candles endpoint

    See /candles as https://tradingstrategy.ai/api/explorer/#/Trading%20pair/web_candles
    ."""
    price = "price"
    tvl = "tvl"


class APIError(Exception):
    """API error parent class."""


class DataNotAvailable(APIError):
    """Data not available.

    This may happen e.g. when a new entry has just come online,
    it has been added to the pair or reserve map, but does not have candles available yet.

    Wraps 404 error from the dataset server.local
    """


class CacheStatus(enum.Enum):
    """When reading cached files, report to the caller about the caching status."""
    cached = "cached"
    cached_with_timestamped_name = "cached_with_timestamped_name"
    missing = "missing"
    expired = "expired"

    def is_readable(self):
        return self in (CacheStatus.cached, CacheStatus.cached_with_timestamped_name)


class CachedHTTPTransport:
    """A HTTP API transport that offers optional local caching of the results.

    - Download live and cached datasets from the candle server and cache locally
      on the filesystem

    - The download files are very large and expect to need several gigabytes of space for them

    - Has a default HTTP retry policy in the case network or server flakiness

    """

    def __init__(self,
                 download_func: Callable,
                 endpoint: Optional[str] = None,
                 cache_period =datetime.timedelta(days=3),
                 cache_path: Optional[str] = None,
                 api_key: Optional[str] = None,
                 timeout: float = 15.0,
                 add_exception_hook=True,
                 retry_policy: Optional[Retry] = None):
        """
        :param download_func: Interactive download progress bar displayed during the download
        :param endpoint: API server we are using - default is `https://tradingstrategy.ai/api`
        :param cache_period: How many days we store the downloaded files
        :param cache_path: Where we store the downloaded files
        :param api_key: Trading Strategy API key to use download
        :param timeout: requests HTTP lib timeout
        :param add_exception_hook: Automatically raise an error in the case of HTTP error. Prevents auto retries.
        :param retry_policy:

            How to handle failed HTTP requests.
            If not given use the default somewhat graceful retry policy.
        """

        self.download_func = download_func

        if endpoint:
            self.endpoint = endpoint
        else:
            self.endpoint = "https://tradingstrategy.ai/api"

        self.cache_period = cache_period

        if cache_path:
            self.cache_path = cache_path
        else:
            self.cache_path = os.path.expanduser("~/.cache/tradingstrategy")

        self.requests = self.create_requests_client(
            api_key=api_key,
            retry_policy=retry_policy,
            add_exception_hook=add_exception_hook,
        )

        self.api_key = api_key
        self.timeout = timeout

    def close(self):
        """Release any underlying sockets."""
        self.requests.close()

    def create_requests_client(self,
                               retry_policy: Optional[Retry]=None,
                               api_key: Optional[str] = None,
                               add_exception_hook=True):
        """Create HTTP 1.1 keep-alive connection to the server with optional authorization details.

        :param add_exception_hook: Automatically raise an error in the case of HTTP error
        """

        session = requests.Session()

        # Set up dealing with network connectivity flakey
        if retry_policy is None:
            # https://stackoverflow.com/a/35504626/315168
            retry_policy = Retry(
                total=5,
                backoff_factor=0.1,
                status_forcelist=[ 500, 502, 503, 504 ],
            )
            session.mount('http://', HTTPAdapter(max_retries=retry_policy))
            session.mount('https://', HTTPAdapter(max_retries=retry_policy))

        if api_key:
            session.headers.update({'Authorization': api_key})

        # - Add default HTTP request retry policy to the client
        package_version = version("trading-strategy")
        system = platform.system()
        release = platform.release()
        session.headers.update({"User-Agent": f"trading-strategy {package_version} on {system} {release}"})

        if add_exception_hook:
            def exception_hook(response: Response, *args, **kwargs):
                if response.status_code == 404:
                    raise DataNotAvailable(f"Server error reply: code:{response.status_code} message:{response.text}")
                elif response.status_code >= 400:
                    raise APIError(f"Server error reply: code:{response.status_code} message:{response.text}")

            session.hooks = {
                "response": exception_hook,
            }
        return session

    def get_abs_cache_path(self):
        return os.path.abspath(self.cache_path)

    def get_cached_file_path(self, fname):
        path = os.path.join(self.get_abs_cache_path(), fname)
        return path

    def get_cached_item(self, fname: Union[str, pathlib.Path]) -> Optional[pathlib.Path]:
        """Get a cached file.

        - Return ``None`` if the cache has expired

        - The cache timeout is coded in the file modified
          timestamp (mtime)
        """

        path = self.get_cached_file_path(fname)
        if not os.path.exists(path):
            # Cached item not yet created
            return None

        f = pathlib.Path(path)

        end_time_pattern = r"-to_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
        if re.search(end_time_pattern, str(fname)):
            # Candle files with an end time never expire, as the history does not change
            return f

        mtime = datetime.datetime.fromtimestamp(f.stat().st_mtime)
        if datetime.datetime.now() - mtime > self.cache_period:
            # File cache expired
            return None

        return f

    def get_cached_item_with_status(
            self,
            fname: Union[str, pathlib.Path]
    ) -> Tuple[pathlib.Path | None, CacheStatus]:
        """Get a cached file.

        - Return ``None`` if the cache has expired

        - The cache timeout is coded in the file modified
          timestamp (mtime)
        """

        path = self.get_cached_file_path(fname)
        if not os.path.exists(path):
            # Cached item not yet created
            return None, CacheStatus.missing

        f = pathlib.Path(path)

        # For some datasets, we encode the end-tie in the fname
        end_time_pattern = r"-to_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
        if re.search(end_time_pattern, str(fname)):
            # Candle files with an end time never expire, as the history does not change
            return f, CacheStatus.cached_with_timestamped_name

        mtime = datetime.datetime.fromtimestamp(f.stat().st_mtime)
        if datetime.datetime.now() - mtime > self.cache_period:
            # File cache expired
            return None, CacheStatus.expired

        return f, CacheStatus.cached

    def _generate_cache_name(
        self,
        pair_ids: Collection[PrimaryKey],
        time_bucket: TimeBucket,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        max_bytes: Optional[int] = None,
        candle_type: str = "candles",
    ) -> str:
        """Generate the name of the file for holding cached candle data for ``pair_ids``.
        """
        # Meaningfully truncate the timestamp to align with the target time bucket.
        if end_time:
            candle_width = time_bucket.to_timedelta()
            trunc = {"second": 0}
            if candle_width >= datetime.timedelta(hours=1):
                trunc["minute"] = 0
            if candle_width >= datetime.timedelta(days=1):
                trunc["hour"] = 0
            end_time = end_time.replace(**trunc)

        # Create a compressed cache key for the filename,
        # as we have 256 char limit on fname lenghts
        full_cache_key = (
            f"{pair_ids}{time_bucket}{start_time}{end_time}{max_bytes}"
        )
        md5 = hashlib.md5(full_cache_key.encode("utf-8")).hexdigest()

        # If exists, include the end time info in filename for cache invalidation logic.
        if start_time:
            start_part = start_time.strftime("%Y-%m-%d_%H-%M-%S")
        else:
            start_part = "any"

        end_part = end_time.strftime("%Y-%m-%d_%H-%M-%S") if end_time else "any"

        return f"{candle_type.replace('_', '-')}-jsonl-{time_bucket.value}-between-{start_part}-and-{end_part}-{md5}.parquet"

    def purge_cache(self, filename: Optional[Union[str, pathlib.Path]] = None):
        """Delete all cached files on the filesystem.

        :param filename:
            If given, remove only that specific file, otherwise clear all cached data.
        """
        target_path = self.cache_path if filename is None else filename

        logger.info("Purging caches at %s", target_path)
        try:
            if os.path.isdir(target_path):
                shutil.rmtree(target_path)
            else:
                os.remove(target_path)
        except FileNotFoundError as exc:
            logger.warning(
                f"Attempted to purge caches, but no such file or directory: {exc.filename}"
            )

    def save_response(self, fpath, api_path, params=None, human_readable_hint: Optional[str]=None):
        """Download a file to the cache and display a pretty progress bar while doing it.

        :param fpath:
            File system path where the download will be saved

        :param api_path:
            Which Trading Strategy backtesting API we call to download the dataset.

        :param params:
            HTTP request params, like the `Authorization` header

        :param human_readable_hint:
            The status text displayed on the progress bar what's being downloaded
        """
        os.makedirs(self.get_abs_cache_path(), exist_ok=True)
        url = f"{self.endpoint}/{api_path}"
        logger.debug("Saving %s to %s", url, fpath)
        # https://stackoverflow.com/a/14114741/315168
        self.download_func(self.requests, fpath, url, params, self.timeout, human_readable_hint)

    def get_json_response(self, api_path, params=None) -> dict:
        url = f"{self.endpoint}/{api_path}"
        logger.debug("get_json_response() %s, %s", url, params)
        response = self.requests.get(url, params=params)
        assert 200 <= response.status_code <= 299
        return response.json()

    def post_json_response(self, api_path, params=None):
        url = f"{self.endpoint}/{api_path}"
        response = self.requests.post(url, params=params)
        return response.json()

    def fetch_chain_status(self, chain_id: int) -> dict:
        """Not cached."""
        return self.get_json_response("chain-status", params={"chain_id": chain_id})

    def fetch_top_pairs(
        self,
        limit: int,
        chain_ids: Collection[ChainId],
        exchange_slugs: Collection[str],
    ) -> dict:
        """Not cached."""
        params = {
            "chain_slugs": ",".join([c.get_slug() for c in chain_ids]),
            "exchange_slugs": ",".join([e for e in exchange_slugs]),
            "limit": str(limit),
        }
        resp = self.get_json_response("top", params=params)
        return resp

    def fetch_pair_universe(self) -> pathlib.Path:
        fname = "pair-universe.parquet"
        cached = self.get_cached_item(fname)

        # Download save the file
        path = self.get_cached_file_path(fname)

        with wait_other_writers(path):

            if cached:
                return cached

            self.save_response(path, "pair-universe", human_readable_hint="Downloading trading pair dataset")
            return self.get_cached_item(fname)

    def fetch_exchange_universe(self) -> pathlib.Path:
        fname = "exchange-universe.json"

        # Download save the file
        path = self.get_cached_file_path(fname)

        with wait_other_writers(path):

            cached = self.get_cached_item(fname)
            if cached:
                return cached

            self.save_response(path, "exchange-universe", human_readable_hint="Downloading exchange dataset")

            _check_good_json(path, "fetch_exchange_universe() failed")

            return self.get_cached_item(fname)
    
    def fetch_lending_reserve_universe(self) -> pathlib.Path:
        fname = "lending-reserve-universe.json"
        cached = self.get_cached_item(fname)

        # Download save the file
        path = self.get_cached_file_path(fname)

        with wait_other_writers(path):

            if cached:
                return cached

            self.save_response(
                path,
                "lending-reserve-universe",
                human_readable_hint="Downloading lending reserve dataset"
            )
            path, status = self.get_cached_item_with_status(fname)

            _check_good_json(path, "fetch_lending_reserve_universe() failed")

            assert status.is_readable(), f"Got status {status} for path"
            return path

    def fetch_candles_all_time(self, bucket: TimeBucket) -> pathlib.Path:
        """Load candles and return a cached file where they are stored.

        - If cached file exists return it directly

        - Wait if someone else is writing the file
          (in multiple parallel testers)
        """
        assert isinstance(bucket, TimeBucket)
        fname = f"candles-{bucket.value}.parquet"
        cached_path = self.get_cached_file_path(fname)

        with wait_other_writers(cached_path):

            cached = self.get_cached_item(fname)
            if cached:
                # Cache exists and is not expired
                return cached

            # Download save the file
            params = {"bucket": bucket.value}
            self.save_response(cached_path, "candles-all", params, human_readable_hint=f"Downloading OHLCV data for {bucket.value} time bucket")
            logger.info(
                "Saved %s as with params %s, down",
                cached_path,
                params
            )
            saved, status = self.get_cached_item_with_status(fname)
            # Troubleshoot multiple test workers race condition
            assert status.is_readable(), f"Cache status {status} with save_response() generated for {fname}, cached path is {cached_path}, download_func is {self.download_func}"
            return saved

    def fetch_liquidity_all_time(self, bucket: TimeBucket) -> pathlib.Path:
        fname = f"liquidity-samples-{bucket.value}.parquet"
        path = self.get_cached_file_path(fname)

        with wait_other_writers(path):

            cached = self.get_cached_item(fname)
            if cached:
                return cached
            # Download save the file
            self.save_response(path, "liquidity-all", params={"bucket": bucket.value}, human_readable_hint=f"Downloading liquidity data for {bucket.value} time bucket")
            return self.get_cached_item(path)

    def fetch_lending_reserves_all_time(self) -> pathlib.Path:
        fname = "lending-reserves-all.parquet"

        # Download save the file
        path = self.get_cached_file_path(fname)

        with wait_other_writers(path):

            cached = self.get_cached_item(fname)
            if cached:
                return cached

            # We only have Aave v3 data for now...
            self.save_response(
                path,
                "aave-v3-all",
                human_readable_hint="Downloading Aave v3 reserve dataset",
            )
            assert os.path.exists(path)
            item, status = self.get_cached_item_with_status(path)
            assert status.is_readable(), f"File not readable after save cached:{cached} fname:{fname} path:{path}"
            return item
    
    def fetch_lending_candles_by_reserve_id(
        self,
        reserve_id: int,
        time_bucket: TimeBucket,
        candle_type: LendingCandleType = LendingCandleType.variable_borrow_apr,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
    ) -> pd.DataFrame:
        """Load particular set of the lending candles and cache the result.

        For the candles format see :py:mod:`tradingstrategy.lending`.

        :param reserve_id:
            Lending reserve's internal id we query data for.
            Get internal id from lending reserve universe dataset.

        :param time_bucket:
            Candle time frame.

        :param candle_type:
            Lending candle type.

        :param start_time:
            All candles after this.
            If not given start from genesis.

        :param end_time:
            All candles before this

        :return:
            Lending candles dataframe
        """

        assert  isinstance(time_bucket, TimeBucket)
        assert isinstance(candle_type, LendingCandleType)

        cache_fname = self._generate_cache_name(
            reserve_id,
            time_bucket,
            start_time,
            end_time,
            candle_type=candle_type.name,
        )

        full_fname = self.get_cached_file_path(cache_fname)

        with wait_other_writers(full_fname):

            cached = self.get_cached_item(cache_fname)

            if cached:
                logger.debug("Using cached data file %s", full_fname)
                return pandas.read_parquet(cached)

            api_url = f"{self.endpoint}/lending-reserve/candles"

            params = {
                "reserve_id": reserve_id,
                "time_bucket": time_bucket.value,
                "candle_types": candle_type,
            }

            if start_time:
                params["start"] = start_time.isoformat()

            if end_time:
                params["end"] = end_time.isoformat()

            try:
                resp = self.requests.get(api_url, params=params, stream=True)
            except DataNotAvailable as e:
                # We have special request hook that translates 404 to this exception
                raise DataNotAvailable(f"Could not fetch lending candles for {params}") from e
            except Exception as e:
                raise APIError(f"Could not fetch lending candles for {params}") from e

            # TODO: handle error
            candles = resp.json()[candle_type]

            df = LendingCandle.convert_web_candles_to_dataframe(candles)

            # Update cache
            path = self.get_cached_file_path(cache_fname)
            df.to_parquet(path)

            size = pathlib.Path(path).stat().st_size
            logger.debug(f"Wrote {cache_fname}, disk size is {size:,}b")

            return df

    def ping(self) -> dict:
        reply = self.get_json_response("ping")
        return reply

    def message_of_the_day(self) -> dict:
        reply = self.get_json_response("message-of-the-day")
        return reply

    def register(self, first_name, last_name, email) -> dict:
        """Makes a register request.

        The request does not load any useful payload, but it is assumed the email message gets verified
        and the user gets the API from the email.
        """
        reply = self.post_json_response("register", params={"first_name": first_name, "last_name": last_name, "email": email})
        return reply

    def fetch_candles_by_pair_ids(
        self,
        pair_ids: Collection[PrimaryKey],
        time_bucket: TimeBucket,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        max_bytes: Optional[int] = None,
        progress_bar_description: Optional[str] = None,
    ) -> pd.DataFrame:
        """Load particular set of the candles and cache the result.

        If there is no cached result, load using JSONL.

        More information in :py:mod:`tradingstrategy.transport.jsonl`.

        For the candles format see :py:mod:`tradingstrategy.candle`.

        :param pair_ids:
            Trading pairs internal ids we query data for.
            Get internal ids from pair dataset.

        :param time_bucket:
            Candle time frame

        :param start_time:
            All candles after this.
            If not given start from genesis.

        :param end_time:
            All candles before this

        :param max_bytes:
            Limit the streaming response size

        :param progress_bar_description:
            Display on downlood progress bar

        :return:
            Candles dataframe
        """
        cache_fname = self._generate_cache_name(
            pair_ids, time_bucket, start_time, end_time, max_bytes
        )

        full_fname = self.get_cached_file_path(cache_fname)

        with wait_other_writers(full_fname):

            cached = self.get_cached_item(cache_fname)

            if cached:
                logger.debug("Using cached JSONL data file %s", full_fname)
                return pandas.read_parquet(cached)

            df: pd.DataFrame = load_candles_jsonl(
                self.requests,
                self.endpoint,
                pair_ids,
                time_bucket,
                start_time,
                end_time,
                max_bytes=max_bytes,
                progress_bar_description=progress_bar_description,
                # temp increase sanity check count
                sanity_check_count=120,
            )

            # Update cache
            path = self.get_cached_file_path(cache_fname)
            df.to_parquet(path)

            size = pathlib.Path(path).stat().st_size
            logger.debug(f"Wrote {cache_fname}, disk size is {size:,}b")

            return df

    def _fetch_tvl_by_pair_id(
        self,
        pair_id: PrimaryKey,
        time_bucket: TimeBucket,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
    ) -> pd.DataFrame:
        """Internal hack to load TVL data for a single pair.

        - See :py:meth:`fetch_tvl_by_pair_ids` for public API.

        - TODO: Currently there is no JSONL endpoint to get liquidity data streaming.
          Thus, we do this hack here. Also because of different data format for Uni v2 and Uni v3,
          the server cannot handle fetching mixed pool types once.

        - This function causes double caching and such.

        - Will be replaced by a proper JSONL streaming in somepoint.

        """

        pair_ids = [pair_id]

        cache_fname = self._generate_cache_name(
            pair_ids,
            time_bucket,
            start_time,
            end_time,
            candle_type="tvl",
        )

        full_fname = self.get_cached_file_path(cache_fname)

        with wait_other_writers(full_fname):

            cached = self.get_cached_item(cache_fname)
            path = self.get_cached_file_path(cache_fname)

            if cached:
                # We have a locally cached version
                logger.debug("Using cached Parquet data file %s", full_fname)
                df = pandas.read_parquet(cached)
            else:
                # Read from the server, store in the disk
                params = {
                    "time_bucket": time_bucket.value,
                    # "pair_ids": ",".join([str(i) for i in pair_ids]),  # OpenAPI comma delimited array
                    "pair_id": pair_ids[0],
                }

                if start_time:
                    params["start"] = start_time.isoformat()

                if end_time:
                    params["end"] = end_time.isoformat()

                params["candle_type"] = OHLCVCandleType.tvl.value

                # Use /candles endpoint to load TVL data
                pair_candle_map = self.get_json_response(
                    "candles",
                    params=params,
                )

                for pair_id, array in pair_candle_map.items():
                    df = XYLiquidity.convert_web_candles_to_dataframe(array)
                    # Fill in pair id,
                    # because /candles endpoint does not reflect it back
                    df["pair_id"] = int(pair_id)

                # Update cache
                df.to_parquet(path)

            size = pathlib.Path(path).stat().st_size
            if not cached:
                logger.debug(f"Wrote {cache_fname}, disk size is {size:,}b")
            else:
                logger.debug(f"Read {cache_fname}, disk size is {size:,}b")

            # Expose if this reply was cached or not,
            # for unit testing
            df.attrs["cached"] = cached
            df.attrs["disk_size"] = size

        return df

    def fetch_tvl_by_pair_ids(
        self,
        pair_ids: Collection[PrimaryKey],
        time_bucket: TimeBucket,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        progress_bar_description: Optional[str] = None,
    ) -> pd.DataFrame:
        """Load particular set of the TVL candles and cache the result.

        For the candles format see :py:mod:`tradingstrategy.liquidity`.

        :param pair_ids:
            Trading pairs internal ids we query data for.
            Get internal ids from pair dataset.

            We should be able to handle unlimited pair count,
            as we do one request per pair.

        :param time_bucket:
            Candle time frame

        :param start_time:
            All candles after this.
            If not given start from genesis.

        :param end_time:
            All candles before this

        :param progress_bar_description:
            Display on downlood progress bar

        :return:
            Liquidity dataframe.

            See :py:mod:`tradingstrategy.liquidity`.
        """

        #
        # TODO: Currently there is no JSONL
        # endpoint to get liquidity data streaming.
        # Thus, we do this hack here.
        # Also because of different data format for Uni v2 and Uni v3,
        # the server cannot handle fetching mixed pool types
        # once.
        #

        chunks = []

        if progress_bar_description:
            # The server does not know the reply size,
            # so we cannot render a progress bar estimation
            progress_bar = tqdm(desc=progress_bar_description, total=len(pair_ids))
        else:
            progress_bar = None

        for pair_id in pair_ids:

            df = self._fetch_tvl_by_pair_id(
                pair_id,
                time_bucket,
                start_time,
                end_time,
            )

            # Work around pd.concat() problem for some reason fails on Github,
            # see details below
            df = df.set_index(["pair_id", "timestamp"], drop=False)

            chunks.append(df)

            if progress_bar:
                progress_bar.update()

        if progress_bar:
            progress_bar.close()

        try:
            return pd.concat(chunks)
        except ValueError as e:
            # Happens only on Github CI
            # https://stackoverflow.com/questions/27719407/pandas-concat-valueerror-shape-of-passed-values-is-blah-indices-imply-blah2
            msg = ""
            for c in chunks:
                msg += f"Index: {c.index}\n"
                msg += f"Data: {c}\n"
            raise ValueError(f"pd.concat() failed:\n{msg}") from e

    def fetch_clmm_liquidity_provision_candles_by_pair_ids(
        self,
        pair_ids: Collection[PrimaryKey],
        time_bucket: TimeBucket,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        progress_bar_description: Optional[str] = None,
    ) -> pd.DataFrame:
        """Stream CLMM Parquet data from the server.

        For the candles format see :py:mod:`tradingstrategy.clmm`.

        :param pair_ids:
            Trading pairs internal ids we query data for.
            Get internal ids from pair dataset.

            We should be able to handle unlimited pair count,
            as we do one request per pair.

        :param time_bucket:
            Candle time frame

        :param start_time:
            All candles after this.
            If not given start from genesis.

        :param end_time:
            All candles before this

        :param progress_bar_description:
            Display on downlood progress bar

        :return:
            CLMM dataframe.

            See :py:mod:`tradingstrategy.clmm`.
        """

        cache_fname = self._generate_cache_name(
            pair_ids, time_bucket, start_time, end_time,
            candle_type="clmm"
        )

        full_fname = self.get_cached_file_path(cache_fname)

        url = f"{self.endpoint}/clmm-candles"

        with wait_other_writers(full_fname):

            cached = self.get_cached_item(cache_fname)
            path = self.get_cached_file_path(cache_fname)

            if not cached:

                params = {
                    "pair_ids": ",".join([str(i) for i in pair_ids]),  # OpenAPI comma delimited array
                    "time_bucket": time_bucket.value,
                    "format": "parquet",
                }

                if start_time:
                    params["start"] = start_time.isoformat()

                if end_time:
                    params["end"] = end_time.isoformat()

                download_with_tqdm_progress_bar(
                    session=self.requests,
                    path=path,
                    url=url,
                    params=params,
                    timeout=self.timeout,
                    human_readable_hint=progress_bar_description,
                )

                size = pathlib.Path(path).stat().st_size
                logger.debug(f"Wrote {cache_fname}, disk size is {size:,}b")

            else:
                size = pathlib.Path(path).stat().st_size
                logger.debug(f"Reading cached Parquet file {cache_fname}, disk size is {size:,}")

            df = pandas.read_parquet(path)

            # Export cache metadata
            df.attrs["cached"] = cached is not None
            df.attrs["filesize"] = size
            df.attrs["path"] = path
            return df

    def fetch_trading_data_availability(self,
          pair_ids: Collection[PrimaryKey],
          time_bucket: TimeBucket,
        ) -> Dict[PrimaryKey, TradingPairDataAvailability]:
        """Check the trading data availability at oracle's real time market feed endpoint.

        - Trading Strategy oracle uses sparse data format where candles
          with zero trades are not generated. This is better suited
          for illiquid DEX markets with few trades.

        - Because of sparse data format, we do not know if there is a last
          candle available - candle may not be available yet or there might not be trades
          to generate a candle

        - This endpoint allows to check the trading data availability for multiple of trading pairs.

        - This endpoint is public

        :param pair_ids:
            Trading pairs internal ids we query data for.
            Get internal ids from pair dataset.

        :param time_bucket:
            Candle time frame

        :return:
            Map of pairs -> their trading data availability

        """

        params = {
            "pair_ids": ",".join([str(i) for i in pair_ids]),  # OpenAPI comma delimited array
            "time_bucket":  time_bucket.value,
        }

        array = self.get_json_response("trading-pair-data-availability", params=params)

        # Make to typed and deseralise
        def _convert(p: dict) -> TradingPairDataAvailability:
            try:
                return {
                    "chain_id": ChainId(p["chain_id"]),
                    "pair_id": p["pair_id"],
                    "pair_address": p["pair_address"],
                    "last_trade_at": datetime.datetime.fromisoformat(p["last_trade_at"]),
                    "last_candle_at": datetime.datetime.fromisoformat(p["last_candle_at"]),
                    "last_supposed_candle_at": datetime.datetime.fromisoformat(p["last_supposed_candle_at"]),
                }
            except Exception as e:
                raise RuntimeError(f"Failed to convert: {p}") from e

        return {p["pair_id"]: _convert(p) for p in array}


@contextmanager
def wait_other_writers(path: Path | str, timeout=120):
    """Wait other potential writers writing the same file.

    - Work around issues when parallel unit tests and such
      try to write the same file

    Example:

    .. code-block:: python

        import urllib
        import tempfile

        import pytest
        import pandas as pd

        @pytest.fixture()
        def my_cached_test_data_frame() -> pd.DataFrame:

            # Al tests use a cached dataset stored in the /tmp directory
            path = os.path.join(tempfile.gettempdir(), "my_shared_data.parquet")

            with wait_other_writers(path):

                # Read result from the previous writer
                if not path.exists(path):
                    # Download and write to cache
                    urllib.request.urlretrieve("https://example.com", path)

                return pd.read_parquet(path)

    :param path:
        File that is being written

    :param timeout:
        How many seconds wait to acquire the lock file.

        Default 2 minutes.

    :raise filelock.Timeout:
        If the file writer is stuck with the lock.
    """

    if type(path) == str:
        path = Path(path)

    assert isinstance(path, Path), f"Not Path object: {path}"

    assert path.is_absolute(), f"Did not get an absolute path: {path}\n" \
                               f"Please use absolute paths for lock files to prevent polluting the local working directory."

    # If we are writing to a new temp folder, create any parent paths
    os.makedirs(path.parent, exist_ok=True)

    # https://stackoverflow.com/a/60281933/315168
    lock_file = path.parent / (path.name + '.lock')

    lock = FileLock(lock_file, timeout=timeout)

    if lock.is_locked:
        logger.info(
            "Parquet file %s locked for writing, waiting %f seconds",
            path,
            timeout,
        )

    with lock:
        yield


def _check_good_json(path: Path, exception_message: str):
    """Check that server gave us good JSON file.

    - 404, 500, API key errors

    """
    broken_data = False

    # Quick fix to avoid getting hit by API key errors here.
    # TODO: Clean this up properly
    with open(path, "rt", encoding="utf-8") as inp:
        data = inp.read()
        try:
            data = json.loads(data)
            if "error" in data:
                broken_data = True
        except JSONDecodeError as e:
            broken_data = True

    if broken_data:
        os.remove(path)
        raise RuntimeError(f"{exception_message}\nJSON data is: {data}")




"""Wrangle DEX market data to be better suitable for algorithmic trading.

- Wrangle is a process where we "massage" incoming price/liquidity data for the isseus we may have encountered during the data collection

- Common DEX data issues are absurd price high/low spikes due to MEV trades

- We also have some open/close values that are "broken" in a sense that they do not reflect the market price you would be able to trade,
  again likely due to MEV

- See :py:func:`fix_dex_price_data` for fixing

"""
import logging
import datetime
from itertools import islice

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy
import numpy as np

from .time import naive_utcnow
from .forward_fill import forward_fill as _forward_fill
from ..pair import PandasPairUniverse
from ..types import AnyTimestamp

logger = logging.getLogger(__name__)


#: Floating point danger zone for price values
DEFAULT_MIN_MAX_RANGE = (0.00000001, 1_000_000.0)


def fix_bad_wicks(
    df: pd.DataFrame,
    threshold=(0.1, 1.9),
    too_slow_threshold=15,
    verbose=False,
    bad_open_close_threshold: float | None=3.0,
) -> pd.DataFrame:
    """Correct out bad high/low values in OHLC data.

    Applicable for both :term:`OHLCV` price feeds and liquidity feeds.

    On :term:`Uniswap` v2 and compatibles, Bad wicks are caused by e.g. very large flash loan, oracle price manipulation attacks,
    and misbheaving bots.

    This function removes bad high/low values and sets them to open/close if they seem to be wildly out of sample.

    :param threshold:
        How many pct % wicks are allowed through.

        Tuple (low threshold, high threshold) relative to close.

        Default to 50%. A high wick cannot be more than 50% of close.

    :param too_slow_threshold:
        Complain if this takes too long

    :param bad_open_close_threshold:
        How many X open must be above the high to be considered a broken data point.

        The open price will be replaced with high price.

        Do not set for liquidity processing.


    :param verbose:
        Make some debug logging when using the function for manual data diagnostics.

    """

    start = naive_utcnow()

    if len(df) == 0:
        return df

    # Optimised with np.where()
    # https://stackoverflow.com/a/65729035/315168
    if threshold is not None:
        df["high"] = np.where(df["high"] > df["close"] * threshold[1], df["close"], df["high"])
        df["low"] = np.where(df["low"] < df["close"] * threshold[0], df["close"], df["low"])

    # For manual diagnostics tracking down bad trading pair data
    if verbose and bad_open_close_threshold:
        bad_opens = df[df["open"] > df["high"] * bad_open_close_threshold]
        for idx, row in islice(bad_opens.iterrows(), 10):
            logger.warning(
                "Pair id %d, timestamp: %s, open: %s, high: %s, buy volume: %s sell volume: %s, volume: %s",
                row.pair_id,
                row.timestamp,
                row.open,
                row.high,
                row.get("buy_volume"),
                row.get("sell_volume"),
                row.get("volume"),
            )
        logger.warn("Total %d bad open price entries detected", len(bad_opens))

    # Issues in open price values with data point - open cannot be higher than high.
    # Not strickly "wicks" but we fix all data while we are at it.
    if bad_open_close_threshold:
        df["open"] = np.where(df["open"] > df["high"] * bad_open_close_threshold, df["high"], df["open"])
        df["close"] = np.where(df["close"] > df["high"] * bad_open_close_threshold, df["high"], df["close"])

    duration = naive_utcnow() - start

    if duration > datetime.timedelta(seconds=too_slow_threshold):
        logger.warning("Very slow fix_bad_wicks(): %s", duration)

    # The following code chokes
    # mask = (df["high"] > df["close"] * (1+threshold)) | (df["low"] < df["close"] * threshold)
    #df.loc[mask, "high"] = df["close"]
    #df.loc[mask, "low"] = df["close"]
    #df.loc[mask, "wick_filtered"] = True
    return df


def filter_bad_wicks(df: pd.DataFrame, threshold=(0.1, 1.9)) -> pd.DataFrame:
    """Mark the bad wicks.

    On :term:`Uniswap` v2 and compatibles, Bad wicks are caused by e.g. very large flash loan, oracle price manipulation attacks,
    and misbheaving bots.

    This function removes bad high/low values and sets them to open/close if they seem to be wildly out of sample.

    :param threshold:
        How many pct % wicks are allowed through as (low, high) tuple.

        This is a tuple (low threshold, high threshold).

        If low < close * threshold[0] ignore the value.

        If high > close * threshold[0] ignore the value.

    """

    df_matches = df.loc[
        (df["high"] > df["close"] * threshold[1]) | (df["low"] < df["close"] * threshold[0])
    ]

    return df_matches


def remove_zero_candles(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """Remove any candle that has a zero value for OHLC

    :param df: Dataframe that may contain zero candles
    :return: pd.Dataframe
    """
    if len(df) > 0:
        filtered_df = df[(df['open'] != 0) & (df['high'] != 0) & (df['low'] != 0) & (df['close'] != 0)]
        return filtered_df
    return df


def remove_min_max_price(
    df: pd.DataFrame,
    min_max_price: tuple = DEFAULT_MIN_MAX_RANGE,
) -> pd.DataFrame:
    """Remove candles where open value is outside the floating point range detector.

    :param df:
        Dataframe that may contain open or close values for too funny precision.

    :param min_max_price:
        Min and max price allowed for open/close before dropping.

    :return: pd.Dataframe
    """
    if len(df) > 0:
        mask = (df["open"] < min_max_price[0]) | (df["close"] < min_max_price[0]) | \
               (df["open"] > min_max_price[1]) | (df["close"] > min_max_price[1])

        filtered_df = df[~mask]
        return filtered_df
    return df


def _replace_for_groups(group, replacement_dict):
    group_name = group.name
    if group_name in replacement_dict:
        for col, value in replacement_dict[group_name].items():
            group[col] = value
    return group


def fix_prices_in_between_time_frames(
    dfgb: pd.DataFrame | DataFrameGroupBy,
    fix_inbetween_threshold: tuple | None = (-0.99, 5.0),
    pair_id_column="pair_id",
):
    """Fix MEV bots breaking open/price.

    TODO: This needs to be later fixed at the data collection level, MEV transactions masked out.

    Example daily candle with a broken open price:

    .. code-block:: text

        COMP-WETH on Uniswap v2 on Ethereum

        2023-12-10 23:00:00	2023-12-10 23:00:00	56.253001	56.253001	55.948422	55.948422	348.915474	2373.049643	0.0	3.0	0.000000	348.915474	1015	18758972	18759006	NaN
        2023-12-11 00:00:00	2023-12-11 00:00:00	0.363204	55.181369	0.354456	55.181369	180421.431179	2357.352148	1.0	6.0	90162.684736	90258.746443	1015	18759541	18759541	NaN
        2023-12-11 02:00:00	2023-12-11 02:00:00	50.059992	50.612159	50.059992	50.612159	2859.011552	2241.581295	1.0	1.0	628.775186	2230.236366	1015	18759912	18760129	NaN

    `Underlying MEV TX causing the issue <https://etherscan.io/tx/0x1106418384414ed56cd7cbb9fedc66a02d39b663d580abc618f2d387348354ab>`__, ChatGPT formatted:

    .. code-block:: text

        +--------------------------------------+-------------------------------------+-------------------+-------------+-------------------+-------------------------------------+-------------------+-------------+-------------+
        |               Action                 |          Swap Amount (Token 1)      |   Value (Token 1)  |  Token 1    |    For (Token 2)   |          Swap Amount (Token 2)      |   Value (Token 2)  |  Token 2    |  Platform   |
        +--------------------------------------+-------------------------------------+-------------------+-------------+-------------------+-------------------------------------+-------------------+-------------+-------------+
        | Aggregated Swap of 3 Tokens           |                                     |                   |             |                   |                                     |                   |             |             |
        | Swap                                 | 20,096.81048332570788104            | $956,407.21       | COMP        | ETH               | 38.247439949510380362              | $101,436.63       | ETH         | Uniswap V2  |
        | Swap                                 | 65.583941528193262325               | $3,121.14         | COMP        | ETH               | 0.010044352430924858               | $26.64            | ETH         | Uniswap V2  |
        | Swap                                 | 0.010044352430924858                | $26.64            | ETH         | DAI               | 23.587216376218824815              | $23.58            | DAI         | Uniswap V2  |
        | Swap                                 | 65.780693352777842112               | $3,130.50         | COMP        | ETH               | 0.010014399362431825               | $26.56            | ETH         | Uniswap V2  |
        | Swap                                 | 0.010014399362431825                | $26.56            | ETH         | DAI               | 23.516703895990702452              | $23.51            | DAI         | Uniswap V2  |
        | Swap                                 | 65.978035432836175638               | $3,139.89         | COMP        | ETH               | 0.009984535616403163               | $26.48            | ETH         | Uniswap V2  |
        | Swap                                 | 0.009984535616403163                | $26.48            | ETH         | DAI               | 23.446402724807108015              | $23.44            | DAI         | Uniswap V2  |
        | Swap                                 | 66.175969539134684165               | $3,149.31         | COMP        | ETH               | 0.009954760926472085               | $26.40            | ETH         | Uniswap V2  |
        | Swap                                 | 0.009954760926472085                | $26.40            | ETH         | DAI               | 23.376312226345838338              | $23.37            | DAI         | Uniswap V2  |
        | Swap                                 | 5.013744524035853058                | $238.60           | COMP        | ETH               | 0.000751785040201655               | $1.99             | ETH         | Uniswap V2  |
        | Swap                                 | 0.000751785040201655                | $1.99             | ETH         | DAI               | 1.765375649239336674               | $1.76             | DAI         | Uniswap V2  |
        | Swap                                 | 38.247439949510380362               | $101,436.63       | ETH         | COMP              | 20,353.762314725950969279          | $968,635.55       | COMP        | Uniswap V2  |
        | Swap                                 | 0.149285130679667947                | $395.92           | ETH         | COMP              | 4,200                              | $199,878.00       | COMP        | Sushiswap   |
        +--------------------------------------+-------------------------------------+-------------------+-------------+-------------------+-------------------------------------+-------------------+-------------+-------------+

    :param dfgb:
        Assume grouped by pair_id and MultiLevel index (pair_id, timestamp).

    """
    assert isinstance(dfgb, DataFrameGroupBy), f"Currently only implemented for DataFrameGroupBy"

    replacements = {}

    for pair_id, price_df in dfgb:
        for column in ("open", "close",):
            price_series = price_df[column]
            fixed_price_series = examine_price_between_time_anomalies(
                price_series,
                low_diff=fix_inbetween_threshold[0],
                high_diff=fix_inbetween_threshold[1],
                heal=True,
            )
            if fixed_price_series is not None:
                replacements[pair_id] = replacements.get(pair_id, {})
                replacements[pair_id][column] = fixed_price_series

    healed = dfgb.apply(lambda x: _replace_for_groups(x, replacements))
    healed = healed.set_index("timestamp", drop=False)
    return healed.groupby(pair_id_column)


def fix_dex_price_data(
    df: pd.DataFrame | DataFrameGroupBy,
    freq: pd.DateOffset | str | None = None,
    forward_fill: bool = True,
    bad_open_close_threshold: float | None = 3.0,
    fix_wick_threshold: tuple | None = (0.1, 1.9),
    fix_inbetween_threshold: tuple | None = (-0.99, 5.0),
    min_max_price: tuple | None = DEFAULT_MIN_MAX_RANGE,
    remove_candles_with_zero: bool = True,
    pair_id_column="pair_id",
    forward_fill_until: AnyTimestamp | None = None,
) -> pd.DataFrame:
    """Wrangle DEX price data for all known issues.

    - Fix broken open/high/low/close value so that they are less likely to cause problems for algorithms

    - Wrangle is a process where we massage incoming price/liquidity data for the isseus we may have encountered during the data collection

    - Common DEX data issues are absurd price high/low spikes due to MEV trades

    - We also have some open/close values that are "broken" in a sense that they do not reflect the market price you would be able to trade,
      again likely due to MEV

    Example:

    .. code-block:: python

          # After we know pair ids that fill the liquidity criteria,
          # we can build OHLCV dataset for these pairs
          print(f"Downloading/opening OHLCV dataset {time_bucket}")
          price_df = client.fetch_all_candles(time_bucket).to_pandas()
          print(f"Filtering out {len(top_liquid_pair_ids)} pairs")
          price_df = price_df.loc[price_df.pair_id.isin(top_liquid_pair_ids)]

          print("Wrangling DEX price data")
          price_df = price_df.set_index("timestamp", drop=False).groupby("pair_id")
          price_df = fix_dex_price_data(
              price_df,
              freq=time_bucket.to_frequency(),
              forward_fill=True,
          )

          print(f"Retrofitting OHLCV columns for human readability")
          price_df = price_df.obj
          price_df["pair_id"] = price_df.index.get_level_values(0)
          price_df["ticker"] = price_df.apply(lambda row: make_full_ticker(pair_metadata[row.pair_id]), axis=1)
          price_df["link"] = price_df.apply(lambda row: make_link(pair_metadata[row.pair_id]), axis=1)

          # Export data, make sure we got columns in an order we want
          print(f"Writing OHLCV CSV")
          del price_df["timestamp"]
          del price_df["pair_id"]
          price_df = price_df.reset_index()
          column_order = ('ticker', 'timestamp', 'open', 'high', 'low', 'close', 'volume', 'link', 'pair_id',)
          price_df = price_df.reindex(columns=column_order)  # Sort columns in a specific order
          price_df.to_csv(
            price_output_fname,
          )
          print(f"Wrote {price_output_fname}, {price_output_fname.stat().st_size:,} bytes")

    :param df:
        Price dataframe with OHLCV data.

        May contain columns named open, close, high, low, volume and timestamp.

        For multipair data this must be `DataFrameGroupBy`.

    :param freq:
        The incoming Pandas frequency of the data, e.g. "d" for daily.

        If the incoming data frequency and `freq` parameter do not match, the data is resampled o the given frequency.

    :param fix_wick_threshold:
        Apply abnormal high/low wick fix filter.

        Percent value of maximum allowed high/low wick relative to close.
        By default fix values where low is 90% lower than close and high is 90% higher than close.

        See :py:func:`~tradingstrategy.utils.groupeduniverse.fix_bad_wicks` for more information.

    :param bad_open_close_threshold:
        See :py:func:`fix_bad_wicks`.

    :param pair_id_column:
        The pair/reserve id column name in the dataframe.

    :param remove_candles_with_zero:
        Remove candles with zero values for OHLC.

        To deal with abnormal data.

    :param min_max_price:
        Remove candles where open value is outside the floating point range detector.

        See :py:func:`remove_min_max_price`.

    :param forward_fill:
        Forward-will gaps in the data.

        Forward-filling data will delete any unknown columns,
        see :py:func:`tradingstrategy.utils.forward_fill.forward_fill` details.

    :param forward_fill_until:
        The timestamp which we know the data is valid for.

        If there are price gaps at rarely traded pairs at the end of the (live) OHLCV series,
        we will forward fill the data until this timestamp.

        If not given forward fills until the last trade of the pair.

    :return:
        Fixed data frame.

        If forward fill is used, all other columns outside OHLCV are dropped.
    """

    assert isinstance(df, (pd.DataFrame, DataFrameGroupBy)), f"Got: {df.__class__}"

    if isinstance(df, DataFrameGroupBy):
        raw_df = df.obj
    else:
        raw_df = df

    if min_max_price:
        logger.info("Removing open/close prices outside our min-max tolerance range: %s", min_max_price)
        raw_df = remove_min_max_price(
            raw_df,
            min_max_price=min_max_price,
        )
        raw_df = raw_df.copy()  # Need to mutate in fix_bad_wicks(), cannot be view

    if fix_wick_threshold or bad_open_close_threshold:
        logger.info("Fixing bad wicks, fix_wick_threshold:%s, bad_open_close_threshold: %s", fix_wick_threshold, bad_open_close_threshold)
        raw_df = fix_bad_wicks(
            raw_df,
            fix_wick_threshold,
            bad_open_close_threshold=bad_open_close_threshold,
        )

    if remove_candles_with_zero:
        logger.info("Fixing zero volume candles")
        raw_df = remove_zero_candles(raw_df)

    # For the further cleanup, need to regroup the DataFrame
    if isinstance(df, DataFrameGroupBy):

        logger.info("Regrouping fixed price data")

        # Need to group here
        # TODO: Make this smarter, but how? Read index data in groupby instance?
        assert "timestamp" in raw_df.columns, f"Got {raw_df.columns}"

        regrouped = raw_df.set_index("timestamp", drop=False).groupby(pair_id_column, group_keys=True)

        logger.info("Fixing prices having bad open/close values between timeframes: %s", fix_inbetween_threshold)

        ff_df = fix_prices_in_between_time_frames(
            regrouped,
            fix_inbetween_threshold=fix_inbetween_threshold,
            pair_id_column=pair_id_column,
        )

    else:
        assert not fix_inbetween_threshold, "fix_inbetween_threshold() only works for DataFrameGroupBy input. Set fix_inbetween_threshold == None if you really want to call this"
        ff_df = raw_df

    if forward_fill:

        if isinstance(forward_fill_until, datetime.datetime):
            forward_fill_until = pd.Timestamp(forward_fill_until)

        logger.info("Forward filling price data")
        assert freq, "freq argument must be given if forward_fill=True"
        df = _forward_fill(ff_df, freq, forward_fill_until=forward_fill_until)
        return df
    else:
        return raw_df


def examine_price_between_time_anomalies(
    price_series: pd.Series,
    high_diff=5.00,
    low_diff=-0.99,
    column: str = "close",
    heal=False,
) -> pd.DataFrame | pd.Series:
    """Find bad open/close prices where the open price is very different from the value of previous and next day.

    - Must be done pair-by-pair

    TODO: Work in progress.

    :param price_series:
        Incoming OHLCV or single price series.

    :param high_diff:
        A price between days cannot be higher than this multiplier.

    :param low_diff:
        A price between days cannot be lower than this multiplier.

    :param column:
        Column name to check.

        Only relevant if input is DataFrame.

    :param heal:
        IF set return the healed price data witih avg values replacing the outliers.

    :return:
        Diagnostics results as DataFrame.

        If `heal` is set, return healed price data or None if no healing needed.
    """
    assert isinstance(price_series, (pd.Series, pd.DataFrame)), f"Got: {price_series.__class__}"

    if isinstance(price_series, pd.Series):
        df = pd.DataFrame({
            "price": price_series,
        })
    else:
        df = price_series
        df["price"] = df[column]

    df = df.copy()

    # Calculate surrounding price
    df["price_before"] = df['price'].shift(1)
    df["price_after"] = df['price'].shift(-1)
    df['surrounding_avg'] = (df["price_after"] + df["price_before"]) / 2

    # How much the current price is different from the surrounding avg price
    df["price_diff"] = ((df["surrounding_avg"] - df['price']) / df['price'])

    # Create mask for rows where we detect anomalies
    df["high_anomaly"] = df["price_diff"] > high_diff
    df["low_anomaly"] = df["price_diff"] < low_diff
    df["anomaly"] =  df["low_anomaly"] | df["high_anomaly"]

    if heal:
        # Only execute if we have data to heal
        if df["anomaly"].sum() > 0:  # Count of True values of anomalies
            df["fixed_price"] = np.where(df["anomaly"], df["surrounding_avg"], df["price"])
            return df["fixed_price"]
        else:
            return None
    else:
        # Get the rows that are anomalies
        # Can't have value at start and end,
        # drop NaNs
        df = df.dropna(subset=["surrounding_avg"])

        return df.loc[df["anomaly"]]


def examine_anomalies(
    pair_universe: PandasPairUniverse | None,
    price_df: pd.DataFrame,
    printer=lambda x: print(x),
    max_print=2,
    pair_id_column: str | None="pair_id",
    open_close_max_diff=5.00,
    open_close_min_diff=-0.99,
    between_high_diff: float | None=5.00,
    between_low_diff: float | None=-0.99,
):
    """Check the price dataframe for data issues.

    - Print out to consoles bad rows in the OHLCV candle price data

    Perform

    - Open/close diff check

    - In between timeframes diff check

    TODO: This is a work in progress helper.

    See also:

    - :py:func:`examine_price_between_time_anomalies`

    :param price_df:
        OHLCV data for multiple trading pairs.

        Can be grouped by `pair_id_column`.

    :param pair_id_column:
        Fix column identifies the pair name in the data.

    :param max_print:
        How many entries print per each anomaly check

    :param open_close_max_diff:
        Abnormal price increase X

    :param open_close_min_diff:
        Abnormal price decrease X
    """

    assert isinstance(price_df, pd.DataFrame)

    issues_found = False

    # Find zero prices
    zero_prices = price_df.loc[price_df["open"] <= 0]

    if pair_id_column:
        zero_prices = zero_prices.drop_duplicates(subset=pair_id_column, keep='first')

    if len(zero_prices) > 0:
        printer(f"Total {len(zero_prices)} rows with zero price entry gap")

    for zero_price_entry in zero_prices.iloc[0:max_print].iterrows():
        printer(f"Found zero price entry {zero_price_entry}")
        issues_found = True

    # Find abnormal price jumps within intraday open->close is very different
    open_close_mask = ((price_df["close"] - price_df["open"]) / price_df["open"]) >= open_close_max_diff
    open_close_mask = open_close_mask | (((price_df["close"] - price_df["open"]) / price_df["open"]) <= open_close_min_diff)

    open_close_gap = price_df.loc[open_close_mask]

    if len(open_close_gap) > 0:
        printer(f"Total {len(open_close_gap)} rows with open/close price gap")

    if pair_id_column:
        open_close_gap = open_close_gap.drop_duplicates(subset=pair_id_column, keep='first')

    if len(open_close_gap) > 0:
        for idx, open_close_entry in open_close_gap.iloc[0:max_print].iterrows():
            diff = (open_close_entry["close"] - open_close_entry["open"]) / open_close_entry["open"]
            printer(f"Found abnormal open/close price diff {diff} at\n{open_close_entry}")
            issues_found = True
    else:
        printer("No open/close price extreme diff anomalies found")


    if between_high_diff and between_low_diff:
        grouped = price_df.groupby(pair_id_column)
        anomalies = []

        between_anomaly_count = 0
        for pair, group in grouped:
            for column in ("open", "close"):
                between_anomalies = examine_price_between_time_anomalies(
                    group,
                    high_diff=between_high_diff,
                    low_diff=between_low_diff,
                    column=column,
                )

                between_anomaly_count += len(between_anomalies)

                # Only record last one per pair
                if len(between_anomalies) > 0:
                    row = between_anomalies.iloc[-1]
                    anomalies.append((pair, row))

        if anomalies:
            print(f"Found {between_anomaly_count} price entries that greatly differ from one before and after")
            for entry in anomalies[0:max_print]:
                pair, diagnostics_row = entry
                timestamp = diagnostics_row["timestamp"]
                printer(f"Found abnormal between price on {pair} at {timestamp}:\n{diagnostics_row}")
            issues_found = True
        else:
            printer("No price time frame in-between extreme diff anomalies found")

    if not issues_found:
        printer(f"No data issues found, {len(price_df)} rows analysed")

    return issues_found




"""OHLCV data forward fill.

:term:`Forward fill` missing OHLCV candles in market data feeds.

- Trading Strategy market data feeds are sparse by default,
  to save bandwidth

- DEXes small cap pairs see fewtrades and if there are no trades in a time frame,
  no candle is generated

- Forward-filled data is used on the client side

- We need to forward fill to make price look up, especially for stop losses faster,
  as forward-filled data can do a simple index look up to get a price,
  instead of backwinding to the last available price


"""
import logging
import warnings

from typing import Collection

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy


logger = logging.getLogger(__name__)


def generate_future_filler_data(
    last_valid_row: pd.Series,
    timestamp: pd.Timestamp,
    columns: Collection[str],
):
    """Create a new placeholder OHLCV entry based on the last valid entry."""
    new_row = {}
    last_close = last_valid_row["close"]

    for col in columns:
        match col:
            case "open" | "high" | "low" | "close":
                new_row[col] = last_close
            case "volume":
                new_row[col] = 0
            case "timestamp":
                new_row[col] = timestamp
            case _:
                raise NotImplementedError(f"Unsupported column {col}")

    return new_row


def fill_future_gap(
    df,
    timestamp: pd.Timestamp,
    columns: Collection[str],
    pair_hint: str | None = None,
):
    """Add a virtual OHLCV value at the end of the pair OHLCV data series if there is no real value."""

    assert isinstance(df, pd.DataFrame)
    assert isinstance(df.index, pd.DatetimeIndex), f"Expected DatetimeIndex index, got {type(df.index)}"

    if timestamp not in df.index:
        # Get the latest valid entry before the timestamp
        last_valid_ts = df.index[-1]
        last_valid_entry = df.loc[last_valid_ts]
        data = generate_future_filler_data(
            last_valid_entry, timestamp, columns
        )
        # Create a new row with the timestamp and the last valid entry's values]
        df.loc[timestamp] = data

        logger.debug(
            "Pair %s: Added data end marker at %s, last entry was %s",
            pair_hint,
            timestamp,
            last_valid_ts
        )
    else:
        logger.debug("Pair %s: no need to add data end marker", pair_hint)

    return df


def fill_future_gap_multi_pair(
    grouped_df,
    timestamp: pd.Timestamp,
    columns: Collection[str],
):
    assert isinstance(grouped_df, DataFrameGroupBy)

    def _apply(df):
        df = fill_future_gap(
            df,
            timestamp,
            columns,
            pair_hint=df.name,
        )
        return df

    fixed = grouped_df.apply(_apply)

    # Annoying: sometimes we may have timestamp column, sometimes we don't
    # depends on the caller and is not normalised
    if "timestamp" in fixed.columns:
        del fixed["timestamp"]

    return fixed.reset_index().set_index("timestamp").groupby("pair_id")


def forward_fill(
    single_or_multipair_data: pd.DataFrame | DataFrameGroupBy,
    freq: pd.DateOffset | str,
    columns: Collection[str] = ("open", "high", "low", "close", "volume", "timestamp"),
    drop_other_columns = True,
    forward_fill_until: pd.Timestamp | None = None,
) -> pd.DataFrame:
    """Forward-fill OHLCV data for multiple trading pairs.

    :py:term:`Forward fill` certain candle columns.

    If multiple pairs are given as a `GroupBy`, then the data is filled
    only for the min(pair_timestamp), max(timestamp) - not for the
    range of the all data.

    .. note ::

        `timestamp` and `pair_id` columns will be deleted in this process
         - do not use these columns, but corresponding indexes instead.

    See also

    - :py:func:`tradingstrategy.utils.groupeduniverse.resample_candles`

    - :py:func:`tradingstrategy.utils.groupeduniverse.resample_series`

    Example:

    .. code-block:: python

        import os

        from tradingstrategy.chain import ChainId
        from tradingstrategy.client import Client
        from tradingstrategy.timebucket import TimeBucket
        from tradingstrategy.utils.forward_fill import forward_fill
        from tradingstrategy.utils.groupeduniverse import fix_bad_wicks

        from tradeexecutor.strategy.execution_context import python_script_execution_context
        from tradeexecutor.strategy.trading_strategy_universe import load_all_data
        from tradeexecutor.strategy.universe_model import UniverseOptions

        client = Client.create_jupyter_client()

        chain_id = ChainId.polygon
        time_bucket = TimeBucket.d1
        exchange_slug = "uniswap-v3"

        exchanges = client.fetch_exchange_universe()
        uni = exchanges.get_by_chain_and_slug(ChainId.polygon, exchange_slug)

        dataset = load_all_data(
            client,
            time_frame=TimeBucket.d1,
            execution_context=python_script_execution_context,
            universe_options=UniverseOptions(),
            with_liquidity=False,
        )

        # Filter out pair ids that belong to our target dataset
        pair_universe = dataset.pairs
        pair_ids = pair_universe.loc[pair_universe["exchange_id"] == uni.exchange_id]["pair_id"]
        filtered_df = dataset.candles.loc[dataset.candles["pair_id"].isin(pair_ids)]

        # Forward fill data
        filtered_df = filtered_df.set_index("timestamp")

        # Sanitise price data
        filtered_df = fix_bad_wicks(filtered_df)

        # Make sure there are no gaps in the data
        filtered_df = filtered_df.groupby("pair_id")
        pairs_df = forward_fill(
            filtered_df,
            freq=time_bucket.to_frequency(),
            columns=("open", "high", "low", "close", "volume"),
        )

        # Wrote Parquest file under /tmp
        fpath = f"/tmp/{chain_id.get_slug()}-{exchange_slug}-candles-{time_bucket.value}.parquet"
        flattened_df = pairs_df.obj
        flattened_df = flattened_df.reset_index().set_index("timestamp")  # Get rid of grouping
        flattened_df.to_parquet(fpath)
        print(f"Wrote {fpath} {os.path.getsize(fpath):,} bytes")

    :param single_or_multipair_data:
        Candle data for single or multiple trading pairs

        - GroupBy DataFrame containing candle data for multiple trading pairs
          (grouped by column `pair_id`).

        - Normal DataFrame containing candle data for a single pair

    :param freq:
        The target frequency for the DataFrame.

    :param columns:
        Columns to fill.

        To save memory and speed, only fill the columns you need.
        Usually `open` and `close` are enough and also filled
        by default.

        To get all OHLC data set this to `("open", "high", "low", "close")`.

        If the data has `timestamp` column we fill it with the first value.

    :param drop_other_columns:
        Remove other columns before forward-fill to save memory.

        The resulting DataFrame will only have columns listed in `columns`
        parameter.

        The removed columns include ones like `high` and `low`, but also Trading Strategy specific
        columns like `start_block` and `end_block`. It's unlikely we are going to need
        forward-filled data in these columns.

        .. note ::

            We have no logic for forward filling random columns, only mentioned columns.

    :param forward_fill_until:
        The timestamp which we know the data is valid for.

        If there are price gaps at rarely traded pairs at the end of the (live) OHLCV series,
        we will forward fill the data until this timestamp.

        If not given forward fills until the last trade of the pair.

        The timestamp must match the index timestamp frequency      .

    :return:
        DataFrame where each timestamp has a value set for columns.

        For multi pair data if input is `DataFrameGroupBy` then a similar `DataFrameGroupBy` is
        returned.
    """

    assert isinstance(single_or_multipair_data, (pd.DataFrame, DataFrameGroupBy))
    assert isinstance(freq, (pd.DateOffset, str)), f"Expected pd.DateOffset, got: {freq}"

    original = single_or_multipair_data
    grouped = isinstance(single_or_multipair_data, DataFrameGroupBy)

    # https://www.statology.org/pandas-drop-all-columns-except/
    if drop_other_columns:
        single_or_multipair_data = single_or_multipair_data[list(columns)]

    # Set the end marker if we know when the data should end
    if forward_fill_until is not None:
        assert isinstance(forward_fill_until, pd.Timestamp), f"Got: {type(forward_fill_until)}"

        if grouped:
            single_or_multipair_data = fill_future_gap_multi_pair(single_or_multipair_data, forward_fill_until, columns)
        else:
            single_or_multipair_data = fill_future_gap(single_or_multipair_data, forward_fill_until, columns)

    # Fill missing timestamps with NaN
    # https://stackoverflow.com/a/45620300/315168
    # This will also ungroup the data
    with warnings.catch_warnings():
        # FutureWarning: https://stackoverflow.com/questions/77969964/deprecation-warning-with-groupby-apply
        warnings.simplefilter("ignore")
        single_or_multipair_data = single_or_multipair_data.resample(freq).mean(numeric_only=True)

    if grouped:
        # resample() will set pair_id to NaN
        # fix here
        single_or_multipair_data["pair_id"] = single_or_multipair_data.index.get_level_values('pair_id')

    columns = set(columns)

    # Force columns to be forward filled in a certain order
    # We always need to ffill close column first
    for column in ("close", "open", "high", "low", "volume", "timestamp"):
        if column in columns:
            columns.remove(column)

            match column:
                case "volume":
                    # Sparse volume is 0
                    single_or_multipair_data["volume"] = single_or_multipair_data["volume"].fillna(0.0)
                case "close":
                    # Sparse close is the previous close
                    single_or_multipair_data["close"] = single_or_multipair_data["close"].fillna(method="ffill")
                case "open" | "high" | "low":
                    # Fill open, high, low from the ffill'ed close.
                    single_or_multipair_data[column] = single_or_multipair_data[column].fillna(single_or_multipair_data["close"])
                case "timestamp":

                    if grouped:
                        check_columns = original.obj.columns
                    else:
                        check_columns = original.columns

                    if isinstance(single_or_multipair_data.index, pd.MultiIndex):
                        if "timestamp" in check_columns:
                            # pair_id, timestamp index
                            single_or_multipair_data["timestamp"] = single_or_multipair_data.index.get_level_values(1)
                    elif isinstance(single_or_multipair_data.index, pd.DatetimeIndex):
                        if "timestamp" in check_columns:
                            # timestamp index
                            single_or_multipair_data["timestamp"] = single_or_multipair_data.index
                    else:
                        raise NotImplementedError(f"Unknown column: {column} - forward_fill() does not know how to handle")

    if columns:
        # Unprocessable columns left
        raise NotImplementedError(f"Does not know how to forward fill: {columns}")

    # Regroup by pair, as this was the original data format
    if grouped:
        single_or_multipair_data["timestamp"] = single_or_multipair_data.index.get_level_values('timestamp')
        single_or_multipair_data = single_or_multipair_data.groupby(level="pair_id")

    return single_or_multipair_data



"""Python requests library downloads with a TQDM progress bar."""
import functools
import shutil
from typing import Optional

from requests import Session
from tqdm_loggable.auto import tqdm


def download_with_tqdm_progress_bar(
        session: Session,
        path: str,
        url: str,
        params: dict,
        timeout: float,
        human_readable_hint: Optional[str]):
    """Use tqdm library to raw a graphical progress bar in notebooks for long downloads.

    Autodetects the Python execution environment

    - Displays HTML progress bar in Jupyter notebooks

    - Displays ANSI progress bar in a console

    See :py:meth:`tradingstrategy.transport.cache.CachedHTTPTransport.save_response` for more information.
    """
    # https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests
    # https://stackoverflow.com/questions/42212810/tqdm-in-jupyter-notebook-prints-new-progress-bars-repeatedly

    r = session.get(url, stream=True, allow_redirects=True, params=params, timeout=timeout)
    if r.status_code != 200:
        try:
            r.raise_for_status()  # Will only raise for 4xx codes, so...
            raise RuntimeError(f"Request to {url} returned status code {r.status_code}")
        except Exception as e:
            # Add more context information
            auth_key = session.headers.get("Authorization", "")[0:12]
            raise RuntimeError(f"Failed to do an API call, using API key {auth_key}...") from e

    file_size = int(r.headers.get('Content-Length', 0))

    desc = human_readable_hint or ""

    # Add warning about missing Content-Length header
    desc += "(Unknown total file size)" if file_size == 0 else ""

    r.raw.read = functools.partial(r.raw.read, decode_content=True)  # Decompress if needed
    with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
        with open(path, "wb") as f:
            shutil.copyfileobj(r_raw, f)

    return path



"""Top trading pair queries.

- Data structures for /top end point

- Used for adding new pairs to open ended trading universe in external trading signal processor

- See :py:func:`tradingstrategy.client.Client.fetch_top_pairs` for usage.
"""
import datetime

from dataclasses import dataclass, field

from dataclasses_json import dataclass_json, config
from marshmallow import fields


@dataclass_json
@dataclass(slots=True)
class TopPairData:
    """See open-defi-api.yaml"""

    #: When this entry was queried
    #:
    #: Wall clock UTC time.
    #:
    #: Because the server serialises as ISO, we need special decoder
    #:
    #: https://github.com/lidatong/dataclasses-json?tab=readme-ov-file#Overriding
    #:
    queried_at: datetime.datetime = field(
        metadata=config(
            decoder=datetime.datetime.fromisoformat,
            mm_field=fields.DateTime(format='iso')
        )
    )

    #: Blockchain this pair is on
    chain_id: int

    #: Internal pair primary key (may change)
    pair_id: int

    #: Internal pair exchange id (may change)
    exchange_id: int

    #: Human readable exchange URL slug (may change)
    exchange_slug: str

    #: Smart contract address of pool smart contract.
    #:
    #: Uniswap v2 pair contract address, Uniswap v3 pool contract address.
    #:
    pool_address: str

    #: Human readable base token
    base_token: str

    #: Human readable quote token
    quote_token: str

    #: Pair fee in 0...1, 0.0030 is 30 BPS
    fee: float

    #: Volume over the last 24h
    #:
    #: May not be available due to latency/denormalisation/etc. issues
    #:
    volume_24h_usd: float | None

    #: Last USD TVL (Uniswap v3) or XY Liquidity (Uniswap v2)
    #:
    #: May not be available due to latency/denormalisation/etc. issues
    #:
    tvl_latest_usd: float | None

    #: When TVL measurement was updated.
    #:
    #: How old data are we using.
    #:
    tvl_updated_at: datetime.datetime | None = field(
        metadata=config(
            decoder=datetime.datetime.fromisoformat,
            mm_field=fields.DateTime(format='iso')
        )
    )


    #: When volume measurement was updated
    #:
    #: How old data are we using.
    #:
    volume_updated_at: datetime.datetime | None = field(
        metadata=config(
            decoder=datetime.datetime.fromisoformat,
            mm_field=fields.DateTime(format='iso')
        )
    )


    #: If this pair was excluded from the top pairs, what was the human-readable heuristics reason we did this.
    #:
    #: This allows you to diagnose better why some trading pairs might not end up in the trading universe.
    #:
    exclude_reason: str | None

    #: TokenSniffer data for this token.
    #:
    #: Used in the filtering of scam tokens.
    #:
    #: Not available for all tokens that are filtered out for other reasons.
    #: This is the last check.
    #:
    #: `See more information here <https://web3-ethereum-defi.readthedocs.io/api/token_analysis/_autosummary_token_analysis/eth_defi.token_analysis.tokensniffer.html>`__.
    #:
    token_sniffer_data: dict | None

    def __repr__(self):
        return f"<Pair {self.base_token} - {self.quote_token} on {self.exchange_slug}, address {self.pool_address} - reason {self.exclude_reason}>"

    def get_ticker(self) -> str:
        """Simple marker ticker identifier for this pair."""
        return f"{self.base_token} - {self.quote_token}"

    def get_persistent_string_id(self) -> str:
        """Stable id over long period of time and across different systems."""
        return f"{self.chain_id}-{self.pool_address}"

    @property
    def token_sniffer_score(self) -> int | None:
        """What was the TokenSniffer score for the base token."""

        if self.token_sniffer_data is None:
            return None

        return self.token_sniffer_data["score"]


@dataclass_json
@dataclass(slots=True)
class TopPairsReply:
    """/top endpoint reply.

    - Get a list of trading pairs, both included and excluded

    """

    #: The top list at the point of time the request was made
    included: list[TopPairData]

    #: Tokens that were considered for top list, but excluded for some reason
    #:
    #: They had enough liquidity, but they failed e.g. TokenSniffer scam check,
    #: or had a trading pair for the same base token with better fees, etc.
    #:
    excluded: list[TopPairData]


"""Uniswap V3 CLMM data format for liquidity provision backtesting.

- See :py:class:`tradingstrategy.client.Client.fetch_clmm_liquidity_provision_candles_by_pair_ids` for documentation.

- `Read more about Concentrated Liquidity Market Making <https://tradingstrategy.ai/glossary/clmm>`__
"""
from dataclasses import dataclass

import pyarrow as pa


@dataclass
class CLMM:
    """Define a single CLMM candle.

    - To be used with `Dementer backtesting framework <https://github.com/zelos-alpha/demeter/tree/master/demeter>`__ but will work with others

    TODO: Document columns.
    """

    def get_pyarrow_schema(cls) -> pa.Schema:
        # This schema is based on the original example files in Demeter repo
        schema = pa.schema([
            ("pair_id", pa.int32()),
            ("bucket", pa.timestamp("s")),
            ("open_tick", pa.int32()),
            ("close_tick", pa.int32()),
            ("high_tick", pa.int32()),
            ("low_tick", pa.int32()),
            ("current_liquidity", pa.decimal256(76)),
            ("net_amount0", pa.decimal256(76)),
            ("net_amount1", pa.decimal256(76)),
            ("in_amount0", pa.decimal256(76)),
            ("in_amount1", pa.decimal256(76)),
        ])
        return schema




"""Filtering Uniswap trading pairs for liquidity.

- Used to build a trading pair universe with tradeable pairs of enough liquidity, no survivorship bias

- See :py:func:`build_liquidity_summary` for usage

"""
from collections import Counter
from typing import Collection, Iterable, Tuple

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy

from tradingstrategy.pair import PandasPairUniverse
from tradingstrategy.types import USDollarAmount, PrimaryKey
from tradingstrategy.utils.time import floor_pandas_week


def get_somewhat_realistic_max_liquidity(
    liquidity_df,
    pair_id,
    samples=10,
    broken_liquidity=100_000_000,
) -> float:
    """Get the max liquidity of a trading pair over its history.

    - Get the token by its maximum ever liquidity, so we avoid survivorship bias

    - Instead of picking the absolute top, we pick n top samples
      and choose lowest of those

    - This allows us to avoid data sampling issues when the liquidity value,
      as calculated with the function of price, might have been weird when the token launched

    :param broken_liquidity:
        Cannot have more than 100M USD

    """
    try:
        liquidity_samples = liquidity_df.obj.loc[pair_id]["close"].nlargest(samples)
        sample = min(liquidity_samples)
        if sample > broken_liquidity:
            # Filter out bad data
            return -1
        return sample
    except KeyError:
        # Pair not available, because liquidity data is not there, or zero, or broken
        return -1


def get_liquidity_today(
    liquidity_df,
    pair_id,
    delay=pd.Timedelta(days=30)
) -> float:
    """Get the current liquidity of a trading pair

    :param delay:
        Look back X days.

        To avoid indexer delays.

    :return:
        US dollars
    """

    try:
        timestamp = floor_pandas_week(pd.Timestamp.now() - delay)
        sample = liquidity_df.obj.loc[pair_id]["close"][timestamp]
        return sample
    except KeyError:
        # Pair not available, because liquidity data is not there, or zero, or broken
        return -1


def build_liquidity_summary(
    liquidity_df: pd.DataFrame | DataFrameGroupBy,
    pair_ids: Collection[PrimaryKey] | pd.Series,
    delay=pd.Timedelta(days=21)
) -> tuple[Counter[PrimaryKey, USDollarAmount], Counter[PrimaryKey, USDollarAmount]]:
    """Build a liquidity status of the trading pairs

    - Get the historical max liquidity of a pair, so we can use this for filtering without survivorship bias

    - Get the most recent liquidity (w/delay of few days)

    Example:

    .. code-block:: python

        chain_id = ChainId.ethereum
        time_bucket = TimeBucket.d1  # OHCLV data frequency
        liquidity_time_bucket = TimeBucket.d1  # TVL data for Uniswap v3 is only sampled daily, more fine granular is not needed
        exchange_slugs = {"uniswap-v3", "uniswap-v2", "sushi"}
        exported_top_pair_count = 100
        liquidity_comparison_date = floor_pandas_week(pd.Timestamp.now() - pd.Timedelta(days=7))  # What date we use to select top 100 liquid pairs
        tokensniffer_threshold = 24  # We want our TokenSniffer score to be higher than this for base tokens
        min_liquidity_threshold = 4_000_000  # Prefilter pairs with this liquidity before calling token sniffer
        allowed_pairs_for_token_sniffer = 150  # How many pairs we let to go through TokenSniffer filtering process (even if still above min_liquidity_threshold)

        #
        # Set up output files - use Trading Strategy client's cache folder
        #
        client = Client.create_jupyter_client()
        cache_path = client.transport.cache_path

        #
        # Download - process - save
        #

        print("Downloading/opening exchange dataset")
        exchange_universe = client.fetch_exchange_universe()

        # Resolve uniswap-v3 internal id
        exchanges = [exchange_universe.get_by_chain_and_slug(chain_id, exchange_slug) for exchange_slug in exchange_slugs]
        exchange_ids = [exchange.exchange_id for exchange in exchanges]
        print(f"Exchange {exchange_slugs} ids are {exchange_ids}")

        # We need pair metadata to know which pairs belong to Polygon
        print("Downloading/opening pairs dataset")
        pairs_df = client.fetch_pair_universe().to_pandas()

        our_chain_pairs = filter_pairs_default(
            pairs_df,
            chain_id=chain_id,
            exchange_ids=exchange_ids,
        )
        our_chain_pair_ids = our_chain_pairs["pair_id"]

        print(f"We have data for {len(our_chain_pair_ids)} trading pairs on {fname} set")
        print("Building pair metadata map")
        pairs_df = pairs_df.set_index("pair_id")
        pair_metadata = {pair_id: row for pair_id, row in pairs_df.iterrows()}
        uni_v3_pair_metadata = {pair_id: row for pair_id, row in pairs_df.iterrows() if row["exchange_slug"] == "uniswap-v3"}
        print(f"From this, Uniswap v3 data has {len(uni_v3_pair_metadata)} pairs")

        # Download all liquidity data, extract
        # trading pairs that exceed our prefiltering threshold
        print(f"Downloading/opening TVL/liquidity dataset {liquidity_time_bucket}")
        liquidity_df = client.fetch_all_liquidity_samples(liquidity_time_bucket).to_pandas()
        print(f"Setting up per-pair liquidity filtering, raw liquidity data os {len(liquidity_df)} entries")
        liquidity_df = liquidity_df.loc[liquidity_df.pair_id.isin(our_chain_pair_ids)]
        liquidity_df = liquidity_df.set_index("timestamp").groupby("pair_id")
        print(f"Forward-filling liquidity, before forward-fill the size is {len(liquidity_df)} samples, target frequency is {liquidity_time_bucket.to_frequency()}")
        liquidity_df = forward_fill(liquidity_df, liquidity_time_bucket.to_frequency(), columns=("close",))  # Only daily close liq needed for analysis, don't bother resample other cols

        # Get top liquidity for all of our pairs
        print(f"Filtering out historical liquidity of pairs")
        pair_liquidity_max_historical, pair_liquidity_today = build_liquidity_summary(liquidity_df, our_chain_pair_ids)
        print(f"Chain {chain_id.name} has liquidity data for {len(pair_liquidity_max_historical)} pairs at {liquidity_comparison_date}")

        # Check how many pairs did not have good values for liquidity
        broken_pairs = {pair_id for pair_id, liquidity in pair_liquidity_max_historical.items() if liquidity < 0}
        print(f"Liquidity data is broken for {len(broken_pairs)} trading pairs")

        uniswap_v3_liquidity_pairs = {pair_id for pair_id in pair_liquidity_max_historical.keys() if pair_id in uni_v3_pair_metadata}
        print(f"From this, Uniswap v3 is {len(uniswap_v3_liquidity_pairs)} pairs")
        assert len(uniswap_v3_liquidity_pairs) > 0, "No Uniswap v3 liquidity detected"

        # Remove duplicate pairs
        print("Prefiltering and removing duplicate pairs")
        top_liquid_pairs_filtered = Counter()
        for pair_id, liquidity in pair_liquidity_max_historical.most_common():
            ticker = make_simple_ticker(pair_metadata[pair_id])
            if liquidity < min_liquidity_threshold:
                # Prefilter pairs
                continue
            if ticker in top_liquid_pairs_filtered:
                # This pair is already in the dataset under a different pool
                # with more liquidity
                continue
            top_liquid_pairs_filtered[pair_id] = liquidity

        print(f"After prefilter, we have {len(top_liquid_pairs_filtered):,} pairs left")

    :param liquidity_df:
        Liquidity data. **MUST BE forward filled for no gaps and timestamp indexed.**

        Must be daily/weekly timeframe to include TVL data and match our lookup functions.

    :param pair_ids:
        Pairs we are interested in

    :param delay:
        The time lag to check the "current" today's liquidity.

        Ensure the data is indexed by the time we run this code.

    :return:
        Two counters of historical max liquidity, liquidity today.

        All in USD.

        Pair liquidity value is set to `-1` if the lookup failed (data not available, data contains inrealistic values, etc.)
    """

    assert isinstance(liquidity_df, (pd.DataFrame, DataFrameGroupBy))
    assert isinstance(pair_ids, (pd.Series, list, set, tuple))
    assert len(pair_ids) > 0

    if not isinstance(liquidity_df, DataFrameGroupBy):
        # TODO: This is unlikely to work but let's try be helpful anyway
        liquidity_df = liquidity_df.set_index("timestamp").groupby("pair_id")

    # Get top liquidity for all of our pairs
    pair_liquidity_max_historical = Counter()
    pair_liquidity_today = Counter()
    for pair_id in pair_ids:
        pair_liquidity_max_historical[pair_id] = get_somewhat_realistic_max_liquidity(liquidity_df, pair_id)
        pair_liquidity_today[pair_id] = get_liquidity_today(liquidity_df, pair_id, delay=delay)
    return pair_liquidity_max_historical, pair_liquidity_today


def get_top_liquidity_pairs_by_base_token(
    pair_universe: PandasPairUniverse,
    pair_liquidity_map: Counter,
    good_base_tokens: list[str],
    count: int,
) -> Iterable[Tuple[PrimaryKey, USDollarAmount]]:
    """Get the top liquidity for pairs.

    :param good_base_token:
        Process pairs in the order of this list, one base token add a time.
    """

    # Then get all pairs with this base token
    result_set = []

    assert len(good_base_tokens) > 0
    assert good_base_tokens[0].startswith("0x")

    #
    for base_token_address in good_base_tokens:
        for pair_id,  liquidity in pair_liquidity_map.items():
            pair_metadata = pair_universe.get_pair_by_id(pair_id)
            if pair_metadata.base_token_address == base_token_address:
                result_set.append((pair_id, liquidity))

        if len(result_set) >= count:
            break

    return result_set



"""Aggregated candles.

- Create aggregated price OHLCV and liquidity candles across all available DEX trading pairs

- See :py:func:`aggregate_ohlcv` for usage

"""

from collections import defaultdict
from dataclasses import dataclass
from typing import TypeAlias

import pandas as pd
import numpy as np
from pandas.core.groupby import DataFrameGroupBy

from tradingstrategy.pair import DEXPair, PandasPairUniverse


@dataclass(frozen=True, slots=True)
class AggregateId:

    chain_id: int
    base_token_symbol: str
    base_token_address: str

    def __repr__(self):
        return f"{self.chain_id}-{self.base_token_symbol}-{self.base_token_address}"


#: trading pair -> underlying  aggregated pair ids map
AggregateMap: TypeAlias = dict[AggregateId, set[int]]

ReverseAggregateMap: TypeAlias = dict[int, AggregateId]


def make_aggregate_id(pair: DEXPair) -> AggregateId:
    """For each aggregated pair, identify them

    - Add both human readable symbols and addresses to a string slug

    :return:
        (chain id)-(base token symbol)-(base token address)
    """
    return AggregateId(
        pair.chain_id.value,
        pair.base_token_symbol,
        pair.base_token_address,
    )


def build_aggregate_map(
    pair_universe: PandasPairUniverse,
) -> tuple[AggregateMap, ReverseAggregateMap]:
    """Generate a dict of pairs that trade the same token.

    - return aggregate_id
    """

    aggregates: dict[AggregateId, set[int]]
    aggregates = defaultdict(set)
    reverse_aggregates: ReverseAggregateMap = {}
    for pair in pair_universe.iterate_pairs():
        agg_id = make_aggregate_id(pair)
        aggregates[agg_id].add(pair.pair_id)
        reverse_aggregates[pair.pair_id] = agg_id
    return aggregates, reverse_aggregates


def calculate_volume_weighted_ohlcv(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate volume weighted average prices (vwap) on OHLCV.

    - Assume input has multiple entries for each timestamp (index) separated by column "pair_id"

    - open, high, low, close columns are weighted by volume column

    - volume is the sum of all volume by timestamp

    - liquidity, if in the columns, is the sum of all liquidity by timestamp

    .. code-block:: text

                    pair_id  open  high  low  close  volume  liquidity
        timestamp
        2020-01-01        1   100   100  100    100     500         10
        2020-02-02        1   100   100  100    100     500         10
        2020-01-01        2   110   110  110    110     250         20
        2020-02-02        2   110   110  110    110     250         20
        2020-02-02        3   200   200  200    200    1000         30


    :param df:
        Must have columns pair_id, timestamp, open, high, low, close, volume and optionally liquidity

    :return:
        Aggregated open, high, low, close, volume, liquidity
    """

    assert isinstance(df.index, pd.DatetimeIndex)

    timestamp_agg = df.groupby(level='timestamp')

    # Calculate 0..1 weight for each (pair_id, timestamp) combo
    df["total_volume_in_timestamp"] = timestamp_agg.agg("volume").sum()
    
    # Some days may see zero trading volume 
    # esp. if the blockchain was halted.
    # In this case we do equal weight    
    # TODO: Need to have smarter logic here for multipair scenarios
    placeholder_volume = 1.0
    df["weight"] = df["volume"].replace(0, placeholder_volume) / df["total_volume_in_timestamp"].replace(0, placeholder_volume)
    
    result_df = pd.DataFrame()
    df["open_weighted"] = (df["open"] * df["weight"])
    df["high_weighted"] = (df["high"] * df["weight"])
    df["low_weighted"] = (df["low"] * df["weight"])
    df["close_weighted"] = (df["close"] * df["weight"])

    grouped_2 = df.groupby("timestamp")
    result_df["open"] = grouped_2["open_weighted"].sum()
    result_df["high"] = grouped_2["high_weighted"].sum()
    result_df["low"] = grouped_2["low_weighted"].sum()
    result_df["close"] = grouped_2["close_weighted"].sum()
    result_df["volume"] = grouped_2["volume"].sum()
    if "liquidity" in df.columns:
        result_df["liquidity"] = grouped_2["liquidity"].sum()
    return result_df


def aggregate_ohlcv_across_pairs(
    pair_universe: PandasPairUniverse,
    price_df: pd.DataFrame,
    liquidity_df: DataFrameGroupBy | None = None,
) -> pd.DataFrame:
    """Builds an aggregates dataframe for trading data.

    - Merge all pools e.g. ETH/USDC Uniswap v2, ETH/USDC Uniswap v3 30 BPS and 5 BPS to a single volume data

    - Prices are weighted by volume

    - Currently supports same-chain pairs only

    Example:

    .. code-block:: python

        from tradingstrategy.utils.aggregate_ohlcv import aggregate_ohlcv_across_pairs

        client = persistent_test_client
        exchange_universe = client.fetch_exchange_universe()
        pairs_df = client.fetch_pair_universe().to_pandas()

        # Create filtered exchange and pair data
        exchange = exchange_universe.get_by_chain_and_slug(ChainId.ethereum, "uniswap-v3")

        pair_universe = PandasPairUniverse.create_pair_universe(
                pairs_df,
                [
                    (exchange.chain_id, exchange.exchange_slug, "WETH", "USDC", 0.0005),
                    (exchange.chain_id, exchange.exchange_slug, "WETH", "USDC", 0.0030),
                    (exchange.chain_id, exchange.exchange_slug, "WETH", "USDC", 0.0100)
                ],
            )
        pair_ids = {p.pair_id for p in pair_universe.iterate_pairs()}
        candles_df = client.fetch_candles_by_pair_ids(
            pair_ids,
            TimeBucket.d7,
            start_time=datetime.datetime(2024, 1, 1),
            end_time=datetime.datetime(2024, 3, 1)
        )
        candles_df = candles_df.groupby("pair_id")
        candles_df = forward_fill(candles_df, "W")

        # fetch_all_liquidity_samples() unnecessary heavy here
        # TODO: Change to dynamic fetch method in the future
        liquidity_df = client.fetch_all_liquidity_samples(TimeBucket.d7).to_pandas()
        liquidity_df = liquidity_df.loc[liquidity_df["pair_id"].isin(pair_ids)]
        liquidity_df = liquidity_df.set_index("timestamp").groupby("pair_id")
        liquidity_df = forward_fill(liquidity_df, "W", columns=("close",))  # Only close liquidity column needd

        aggregated_df = aggregate_ohlcv_across_pairs(
            pair_universe,
            candles_df,
            liquidity_df["close"],
        )

    :param pair_universe:
        Pair metadata

    :param price_df:
        OHLCV dataframe.

        Must be forward filled.

    :param liquidity_df:
        Liquidity dataframe.

        Must be forward filled.

        Only "close" column is used.

    :return:
        DataFrame with following colmuns

        - aggregate_id
        - base
        - quote
        - open
        - low
        - high
        - close
        - volume
        - liquidity
        - pair_ids (list of ints)

        Volume and liquidity are in USD.
    """

    assert isinstance(pair_universe, PandasPairUniverse)
    # assert isinstance(price_df, pd.DataFrame)

    aggregates, reverse_aggregates =  build_aggregate_map(
        pair_universe,
    )

    assert "timestamp" in price_df.obj.columns  # TODO: Generate from index if not present as a column
    price_indexed_pair_df = price_df.obj.droplevel("timestamp")

    price_df_raw = price_df.obj
    if liquidity_df is not None:
        liquidity_df_raw = liquidity_df.obj
    else:
        liquidity_df_raw = None

    # result_df = pd.DataFrame()
    chunks = []

    # Aggregate each asset by its underlying trading pairs
    for agg_id, pair_ids in aggregates.items():
        # Select all candle data where trading pair belongs to this aggregate
        selected_rows = price_df_raw.loc[price_df_raw.index.get_level_values(0).isin(pair_ids)]
        selected_rows = selected_rows.copy()  # Need to add to these DF
        selected_rows["pair_id"] = selected_rows.index.get_level_values(0)

        if liquidity_df_raw is not None:
            # pair_id  timestamp
            # 2697585  2023-07-02    2.654539e+07
            #          2023-07-09    1.086443e+07

            liquidity_rows = liquidity_df_raw.loc[liquidity_df_raw.index.get_level_values(0).isin(pair_ids)]
            selected_rows["liquidity"] = liquidity_rows

        selected_rows = selected_rows.set_index("timestamp")
        aggregated_rows = calculate_volume_weighted_ohlcv(selected_rows)
        aggregated_rows["aggregate_id"] = str(agg_id)
        aggregated_rows["base"] = agg_id.base_token_symbol

        # https://stackoverflow.com/a/71977912/315168
        q = np.array([1,], dtype=object)   # dummy array, note the dtype
        q[0] = list(pair_ids)                      # squeeze the list into single cell
        aggregated_rows["pair_ids"] = np.tile(q, aggregated_rows.shape[0])  # tile and assign

        chunks.append(aggregated_rows)

    return pd.concat(chunks)
